;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "ivy-prescient" "20260628.2243"
  "Prescient.el + Ivy."
  '((emacs     "25.1")
    (prescient "6.1.0")
    (ivy       "0.11.0"))
  :url "https://github.com/raxod502/prescient.el"
  :commit "5649977fa7789e4615efeca09397ed7eccd06dfc"
  :revdesc "5649977fa778"
  :keywords '("extensions")
  :authors '(("Radian LLC" . "contact+prescient@radian.codes"))
  :maintainers '(("Radian LLC" . "contact+prescient@radian.codes")))
